/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */

/* 
 * Copyright 2006, Clemson University
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *    1. Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Based on data from:
 * Meng et al, "Characterizing Flows in Large Wireless Data Networks", 
 * MobiCom 2004
 * AND
 * Kotz and Essien, "Analysis of a Campus-wide Wireless Network",
 * Mobicom 2002
 */

#include <tclcl.h>
#include "lib/bsd-list.h"
#include "random.h"
#include "tcp-full.h"
#include "wlantraf-full.h"

/*::::::::::::::::::::::::::::::: WLANTRAF::::::::::::::::::::::::::::::::::*/

static class WLANTrafFullClass : public TclClass {
public:
	WLANTrafFullClass() : TclClass("WLANTrafFull") {}
	TclObject* create(int, const char*const*) {
		return (new WLANTrafFull);
	}
} class_wlantraffull;

WLANTrafFull::WLANTrafFull() :
	TclObject(), timer_(this), connection_interval_(0), 
	BSifq_(NULL),
	client_ind_(0), client_nodes_(0),
	outfp_(NULL), ID_(-1), run_(0), debug_(0),
	incoming_percent_(50), outgoing_percent_(50),
	active_connections_(0), total_connections_(-1), 
	running_(0), flowarrive_rv_(NULL), flowsize_rv_(NULL)
{
	int i;
	strcpy (tcptype_, "Sack");
	server_ = NULL;
	for (i=0; i<MAX_NODES; i++) {
		client_[i] = NULL;
	}
	// zero out response time PDF bins
	clearBins();
}

WLANTrafFull::~WLANTrafFull()
{
	Tcl& tcl = Tcl::instance();

	// output stats
	if (debug_ > 0) {
		fprintf (stderr, "total connections created: %d ", 
			 total_connections_+1);
		fprintf (stderr, "in pool: %d  active: %d\n", 
			 serverAppPool_.size(), serverAppActive_.size());
	}
	
	// delete timer
	timer_.force_cancel();

	// delete active clients in the pool
	map<string, WLANTrafFullClientApp*>::iterator ca_iter;
	for (ca_iter = clientAppActive_.begin(); 
	     ca_iter != clientAppActive_.end(); ca_iter++) {
		ca_iter->second->stop();
		tcl.evalf ("delete %s", ca_iter->second->name());
		clientAppActive_.erase (ca_iter);
	}

	// delete active servers in the pool
	map<string, WLANTrafFullServerApp*>::iterator sa_iter;
	for (sa_iter = serverAppActive_.begin(); 
	     sa_iter != serverAppActive_.end(); sa_iter++) {
		sa_iter->second->stop();
		tcl.evalf ("delete %s", sa_iter->second->name());
		serverAppActive_.erase (sa_iter);
	}

	// delete clients in pool
	WLANTrafFullClientApp* cli;
	while (!clientAppPool_.empty()) {
		cli = clientAppPool_.front();
		cli->stop();
		tcl.evalf ("delete %s", cli->name());
		clientAppPool_.pop();
	}

	// delete servers in pool
	WLANTrafFullServerApp* srv;
	while (!serverAppPool_.empty()) {
		srv = serverAppPool_.front();
		srv->stop();
		tcl.evalf ("delete %s", srv->name());
		serverAppPool_.pop();
	}

	// delete agents in the pool
	FullTcpAgent* tcp;
	while (!tcpPool_.empty()) {
		tcp = tcpPool_.front();
		tcl.evalf ("delete %s", tcp->name());
		tcpPool_.pop();
	}

	// close output file
	if (outfp_)
		fclose(outfp_);
	if (statsfp_)
		fclose(statsfp_);
}

FullTcpAgent* WLANTrafFull::picktcp()
{
	FullTcpAgent* a;
	Tcl& tcl = Tcl::instance();

	if (tcpPool_.empty()) {
		tcl.evalf ("%s alloc-tcp %s", name(), tcptype_);
		a = (FullTcpAgent*) lookup_obj (tcl.result());
		if (a == NULL) {
			fprintf (stderr, "Failed to allocate a TCP agent\n");
			abort();
		}
		if (debug_ > 1) {
			fprintf (stderr, 
				 "\tflow %d created new TCPAgent %s\n",
				 total_connections_, a->name());
		}
	} else {
		a = tcpPool_.front(); 	// grab top of the queue
		tcpPool_.pop();         // remove top from queue

		if (debug_ > 1) {
			fprintf (stderr, "\tflow %d got TCPAgent %s", 
				 total_connections_, a->name());
			fprintf (stderr, " from pool (%d in pool)\n",
				 tcpPool_.size());
		}
	}

	return a;
}

WLANTrafFullServerApp* WLANTrafFull::pickServerApp()
{
	WLANTrafFullServerApp* a;

	if (serverAppPool_.empty()) {
		Tcl& tcl = Tcl::instance();
		tcl.evalf ("%s alloc-server-app", name());
		a = (WLANTrafFullServerApp*) lookup_obj (tcl.result());
		if (a == NULL) {
			fprintf (stderr, 
				"Failed to allocate a WLANTrafFull server app\n");
			abort();
		}
		if (debug_ > 1)
			fprintf (stderr, "\tflow %d created new ",
				 total_connections_);
	} else {
		a = serverAppPool_.front();   // grab top of the queue
		serverAppPool_.pop();         // remove top from queue

		if (debug_ > 1)
			fprintf (stderr, "\tflow %d got ", 
				 total_connections_);
	}

	// initialize server app
	a->set_id(total_connections_);
	a->set_mgr(this);

	if (debug_ > 1)
		fprintf (stderr, "ServerApp %s (%d in pool, %d active)\n",
			 a->name(), serverAppPool_.size(), 
			 serverAppActive_.size()+1);

	return a;
}

WLANTrafFullClientApp* WLANTrafFull::pickClientApp()
{
	WLANTrafFullClientApp* a;

	if (clientAppPool_.empty()) {
		Tcl& tcl = Tcl::instance();
		tcl.evalf ("%s alloc-client-app", name());
		a = (WLANTrafFullClientApp*) lookup_obj (tcl.result());
		if (a == NULL) {
			fprintf (stderr, 
				"Failed to allocate a WLANTrafFull client app\n");
			abort();
		}
		if (debug_ > 1)
			fprintf (stderr, "\tflow %d created new ",
				 total_connections_);
	} else {
		a = clientAppPool_.front();   // grab top of the queue
		clientAppPool_.pop();         // remove top from queue

		if (debug_ > 1)
			fprintf (stderr, "\tflow %d got ", 
				 total_connections_);
	}

	// initialize client app
	a->set_id(total_connections_);
	a->set_mgr(this);

	if (debug_ > 1)
		fprintf (stderr, "ClientApp %s (%d in pool, %d active)\n",
			 a->name(), clientAppPool_.size(), 
			 clientAppActive_.size()+1);

	return a;
}

void WLANTrafFull::recycle(FullTcpAgent* agent)
{
	/*
	 * Problem is that server apps are "done" before client apps,
	 * so server agents and apps get recycled and re-used before
	 * client apps & agents that were attached to them are
	 *
	 */
	if (agent == NULL) {
		fprintf (stderr, "recycle> agent is null\n");
		return;
	}

	// reinitialize FullTcp agent
	agent->reset();

	// add to the inactive agent pool
	tcpPool_.push (agent);

	if (debug_ > 2) {
		fprintf (stderr, "\tTCPAgent %s moved to pool ", 
			 agent->name());
		fprintf (stderr, "(%d in pool)\n", tcpPool_.size());
	}
}

void WLANTrafFull::recycle(WLANTrafFullClientApp* app)
{
	if (app == NULL)
		return;

	// find the client app in the active pool
	map<string, WLANTrafFullClientApp*>::iterator ca_iter = 
		clientAppActive_.find(app->get_agent_name());
	if (ca_iter == clientAppActive_.end()) 
		return;

	// remove the client app from the active pool
	clientAppActive_.erase(ca_iter);

	// insert the client app into the inactive pool
	clientAppPool_.push (app);

	if (debug_ > 2) {
		fprintf (stderr, "\tClientApp %s (%d) moved to pool ", 
			 app->name(), app->get_id());
		fprintf (stderr, "(%d in pool, %d active)\n", 
			 clientAppPool_.size(), clientAppActive_.size());
	}

	// recycle app
	app->recycle();
}

void WLANTrafFull::recycle(WLANTrafFullServerApp* app)
{
	if (app == NULL)
		return;

	// find the server app in the active pool
	map<string, WLANTrafFullServerApp*>::iterator sa_iter = 
		serverAppActive_.find(app->get_agent_name());
	if (sa_iter == serverAppActive_.end()) 
		return;

	// remove the server app from the active pool
	serverAppActive_.erase(sa_iter);

	// insert the server app into the inactive pool
	serverAppPool_.push (app);

	if (debug_ > 2) {
		fprintf (stderr, "\tServerApp %s (%d) moved to pool ",
			 app->name(), app->get_id());
		fprintf (stderr, "(%d in pool, %d active)\n", 
			 serverAppPool_.size(), serverAppActive_.size());
	}

	// recycle app
	app->recycle();
}

double WLANTrafFull::connection_interval() 
{
	return connection_interval_;
}

int WLANTrafFull::get_flowsize() 
{
	return (int) (flowsize_rv_->value());
}

void WLANTrafFull::setup_connection()
/*
 * Setup a new connection, including creation of Agents and Apps
 */
{
	int flowsize, reqsize, rspsize;
	Tcl& tcl = Tcl::instance();

	// incr count of connections
	active_connections_++;
	total_connections_++;

	if (debug_ > 1) {
		fprintf (stderr, 
		 "\nWLANTrafFull %s> new flow %d  total active: %d at %f\n",
		 name(), total_connections_, active_connections_, now());
	}

	// rotate through nodes assigning connections
	client_ind_++;
	if (client_ind_ >= client_nodes_)
		client_ind_ = 0;
	if (debug_ > 1) {
		fprintf (stderr, "\ton node %s (client_[%d])\n", client_[client_ind_]->name(),
			 client_ind_);
	}

	// pick tcp agent for client and server
	FullTcpAgent* ctcp = picktcp();
	FullTcpAgent* stcp = picktcp();

	// attach agents to nodes (server_ client_)
	tcl.evalf ("%s attach %s", server_->name(), 
		   stcp->name());
	tcl.evalf ("%s attach %s", client_[client_ind_]->name(), 
		   ctcp->name());

	// set TCP options
	tcl.evalf ("%s setup-tcp %s %d", name(), stcp->name(), 
		   total_connections_);
	tcl.evalf ("%s setup-tcp %s %d", name(), ctcp->name(),
		   total_connections_);

	// create WLANTrafFullApps
	WLANTrafFullClientApp* client_app = pickClientApp();
	WLANTrafFullServerApp* server_app = pickServerApp();

	// attach TCPs to WLANTrafFullApps
	ctcp->attachApp ((Application*) client_app);
	stcp->attachApp ((Application*) server_app);
	client_app->set_server(server_app);
	client_app->set_agent(ctcp);
	server_app->set_agent(stcp);

	// get flow size
	flowsize = get_flowsize();
	reqsize = (int) (flowsize * (outgoing_percent_ / 100.0));
	rspsize = (int) (flowsize * (incoming_percent_ / 100.0));
	if (reqsize < 1) {
		client_app->set_reqsize(1);
	}
	if (rspsize < 1) {
		client_app->set_rspsize(1);
	}
	client_app->set_reqsize(reqsize);
	client_app->set_rspsize(rspsize);
	server_app->set_reqsize(reqsize);
	server_app->set_rspsize(rspsize);

	if (debug() > 2) {
		fprintf (stderr, "new flow (%d)> flowsize %d  req %d  rsp %d\n", 
			 client_app->get_id(), flowsize, reqsize, rspsize);
	}

	// set priority for the flow
	if (BSifq_ != NULL) {
		// set priorities
		Tcl& tcl = Tcl::instance();
		tcl.evalf ("%s set-priorities %s %d %s %s %d\n", name(), 
			   BSifq_->name(), ac_thresh_, ctcp->name(), 
			   stcp->name(), rspsize);
	}


	// setup connection between client and server
	tcl.evalf ("set ns [Simulator instance]");
	tcl.evalf ("$ns connect %s %s", ctcp->name(), stcp->name());
	tcl.evalf ("%s listen", stcp->name());

	// put apps in active list
	clientAppActive_[ctcp->name()] = client_app;
	serverAppActive_[stcp->name()] = server_app;

	// start WLANTrafFullApps
	client_app->start();
	server_app->start();

	// set time for next connection to start
	connection_interval_ = flowarrive_rv_->value();
}

void WLANTrafFull::start()
{            
	running_ = 1;

	// schedule first connection
	timer_.sched (0);
}

void WLANTrafFull::stop()
{
	running_ = 0;
}

int WLANTrafFull::command(int argc, const char*const* argv) {
	if (argc == 2) {
		if (strcmp (argv[1], "start") == 0) {
			start();
			return (TCL_OK);
		}
		else if (strcmp (argv[1], "stop") == 0) {
			stop();
			return (TCL_OK);
		}
		else if (strcmp (argv[1], "active-connections") == 0) {
			if (outfp_) {
				fprintf (outfp_, "%d ", active_connections_);
				fflush (outfp_);
			}
			else
				fprintf (stderr, "%d ", active_connections_);
			return (TCL_OK);
		}
		else if (strcmp (argv[1], "total-connections") == 0) {
			fprintf (stderr, "%d ", total_connections_);
			return (TCL_OK);
		}
	}
	else if (argc == 3) {
		if (!strcmp (argv[1], "set-client")) {
			if (client_ind_ >= MAX_NODES) {
				return (TCL_ERROR);
			}
			client_[client_ind_] = (Node*) 
				lookup_obj(argv[2]);
			if (client_[client_ind_] == NULL) {
				return (TCL_ERROR);
			}
			client_ind_++;
			client_nodes_++;
			return (TCL_OK);			
		}
		else if (!strcmp (argv[1], "set-server")) {
			server_ = (Node*) lookup_obj(argv[2]);
			if (server_ == NULL)
				return (TCL_ERROR);
			return (TCL_OK);			
		}
		else if (!strcmp (argv[1], "set-BSifq")) {
			BSifq_ = (PriQ*) lookup_obj(argv[2]);
			if (BSifq_ == NULL)
				return (TCL_ERROR);
			return (TCL_OK);
		}
		else if (!strcmp (argv[1], "set-ac-thresh")) {
			ac_thresh_ = atoi(argv[2]);
			return (TCL_OK);
		}
		else if (!strcmp (argv[1], "set-flow_size")) {
                        int res = lookup_rv (flowsize_rv_, argv[2]);
                        if (res == TCL_ERROR) {
                                delete flowsize_rv_;
                                fprintf (stderr, "Invalid flow size ");
                                fprintf (stderr, "random variable\n");
                                return (TCL_ERROR);
                        }
                        return (TCL_OK);
                }
                else if (!strcmp (argv[1], "set-flow_arrive")) {
                        int res = lookup_rv (flowarrive_rv_, argv[2]);
                        if (res == TCL_ERROR) {
                                delete flowarrive_rv_;
                                fprintf (stderr,"Invalid flow arrive ");
                                fprintf (stderr, "random variable\n");
                                return (TCL_ERROR);
                        }
                        return (TCL_OK);
                }
		else if (!strcmp (argv[1], "set-tcp-type")) {
			strcpy (tcptype_, argv[2]);
			return (TCL_OK);
		}
		else if (!strcmp (argv[1], "set-incoming-percent")) {
			incoming_percent_ = atoi(argv[2]);
			return (TCL_OK);
		}
		else if (!strcmp (argv[1], "set-outgoing-percent")) {
			outgoing_percent_ = atoi(argv[2]);
			return (TCL_OK);
		}
		else if (strcmp (argv[1], "set-outfile") == 0) {
			outfp_ = fopen (argv[2], "w");
			if (outfp_)
				return (TCL_OK);
			else 
				return (TCL_ERROR);
		}
		else if (strcmp (argv[1], "set-ID") == 0) {
			ID_ = (int) atoi (argv[2]);
			return (TCL_OK);
		}
		else if (strcmp (argv[1], "set-run") == 0) {
			run_ = (int) atoi (argv[2]);
			return (TCL_OK);
		}
		else if (strcmp (argv[1], "set-debug") == 0) {
			debug_ = (int) atoi (argv[2]);
			return (TCL_OK);
		}
		else if (strcmp (argv[1], "set-stats-outfile") == 0) {
			statsfp_ = fopen (argv[2], "w");
			if (statsfp_)
				return (TCL_OK);
			else 
				return (TCL_ERROR);
		}
		else if (strcmp (argv[1], "recycle") == 0) {
			FullTcpAgent* tcp = (FullTcpAgent*) lookup_obj(argv[2]);
			
			/*
			 * Need to wait to recycle server until client
			 * is done.  Fortunately, client has
			 * a handle to the server.
			 */

			// find client app associated with this agent
			map<string, WLANTrafFullClientApp*>::iterator ca_iter = 
				clientAppActive_.find(tcp->name());
			if (ca_iter == clientAppActive_.end()) {
				// this isn't a client app, but a server app
				return (TCL_OK);
			}

			WLANTrafFullClientApp* ca = ca_iter->second;
			WLANTrafFullServerApp* sa = ca->get_server();
	FullTcpAgent* stcp = (FullTcpAgent*) lookup_obj(sa->get_agent_name());

			if (debug_ > 1)
			      fprintf (stderr, "client %s (%d)> DONE at %f\n",
					 ca->name(), ca->get_id(), now());

			// remove apps from active pools and put 
			// in inactive pools
			recycle ((FullTcpAgent*) tcp);
			recycle ((FullTcpAgent*) stcp);
			recycle (ca);
			recycle (sa);

			active_connections_--;   // one less active conn
			return (TCL_OK);
		}
	}
	return TclObject::command(argc, argv);
}

/*
 * Response Time Bins
 */

void WLANTrafFull::clearBins()
{
	int i;
	for (i=0; i<=RSPTM_BINS; i++) {
		rsptmBins[i] = 0;
	}
	nRspTms = 0;
}

void WLANTrafFull::addToBin(double rsptm)
{
	int bin = (int) (rsptm / RSPTM_BINSZ);  // truncate
	if (bin > RSPTM_BINS) {
		bin = RSPTM_BINS;
	}
	rsptmBins[bin]++;
	nRspTms++;
}

void WLANTrafFull::printBins()
{
	int i;
	for (i=0; i<=RSPTM_BINS; i++) {
		if (rsptmBins[i] == 0) {
			continue;
		}
		if (i < RSPTM_BINS) {
			fprintf (stderr, "rsptmBins[%d] (%d-%d ms): %d\n",
				 i, (i*RSPTM_BINSZ), ((i+1)*RSPTM_BINSZ)-1,
				 rsptmBins[i]);
		}
		else if (i == RSPTM_BINS) {
			fprintf (stderr, "rsptmBins[%d] (over %d ms): %d\n",
				 i, (i*RSPTM_BINSZ), rsptmBins[i]);
		}
	}
}

int WLANTrafFull::getPercentile (double percent)
{
	int bin = 0;           // bin index
	int local_rspTms = 0;  // number of rspTms seen so far

	// position is floor (percent * nRspTms) + 1
	int pos = (int) (percent * nRspTms) + 1;

	// find the bin that has the pos_th entry
	while (local_rspTms < pos && bin <= RSPTM_BINS) {
		local_rspTms += rsptmBins[bin];
		bin++;
	}
	// when the while loop ends, we'll be one past appropriate bin,
	// this will give the upper-bound of the bin where the percent
	// falls -- or 5049 if it's in the last bin
	return ((bin * RSPTM_BINSZ)-1);
}

/*:::::::::::::::::::::::::::: WLANTRAFTIMER ::::::::::::::::::::::::::::::::*/

void WLANTrafFullTimer::expire(Event* = 0) 
{

}

void WLANTrafFullTimer::handle(Event* e)
{
	if (!mgr_->running()) {
		if (mgr_->get_active() != 0) {
			TimerHandler::handle(e);
			sched (1);              // check back in 1 second...
		}
		return;
	}

	// setup new connection
	mgr_->setup_connection();

	TimerHandler::handle(e);

	// schedule time for next connection
	if (mgr_->debug() > 1) {
		fprintf (stderr, "\tnext connection scheduled in %f s...\n", 
			 mgr_->connection_interval());
	}
	sched (mgr_->connection_interval());
}


/*:::::::::::::::::::::::::: WLANTRAFCLIENTAPP ::::::::::::::::::::::::::::::*/

static class WLANTrafFullClientAppClass : public TclClass {
public:
	WLANTrafFullClientAppClass() : TclClass("Application/WLANTrafFull/Client") {}
	TclObject* create(int, const char*const*) {
		return (new WLANTrafFullClientApp);
	}
} class_app_wlantraf_client;

WLANTrafFullClientApp::~WLANTrafFullClientApp()
{
	Tcl& tcl = Tcl::instance();
	if (agent_ != NULL) {
		tcl.evalf ("delete %s", agent_->name());
	}
}

void WLANTrafFullClientApp::start()
{
	running_ = 1;
	send();
}

void WLANTrafFullClientApp::stop()
{
	running_ = 0;
}

void WLANTrafFullClientApp::recycle()
{
	id_ = 0;
	rspsize_ = 0;
	totalbytes_ = 0;
	agent_ = NULL;
	server_ = NULL;
}

void WLANTrafFullClientApp::send()
{
	if (reqsize_ < 1) {
		return;
	}

	if (!running_) {
		return;
	}

	time_of_req_ = mgr_->now();
	
	// send request
	agent_->sendmsg(reqsize_);
	if (mgr_->debug() > 1)
		fprintf (stderr, "client %s (%d)> sent %d-byte req (%d-byte rsp) at %f\n",
			 name(), id_, reqsize_, rspsize_, time_of_req_);
}

void WLANTrafFullClientApp::recv(int bytes)
{
	// we've received a packet
	if (mgr_->debug() > 3)
		fprintf (stderr, "client %s (%d)> received %d bytes at %f\n", 
			 name(), id_, bytes, mgr_->now());

	totalbytes_ += bytes;

	if (totalbytes_ == rspsize_) {
		// we've received all packets from server
		if (mgr_->debug() > 1) {
			fprintf (stderr, "client %s (%d)> received ",
				 name(), id_);
			fprintf (stderr, "total of %d bytes at %f\n", 
				 rspsize_, mgr_->now());
		}
		
		stop();

		// reset totalbytes
		totalbytes_ = 0;

		// update response time PDF
		double now = mgr_->now();
		double rsptm = (now - time_of_req_) * 1000.0;
		mgr_->addToBin(rsptm);

		/* DEBUGGING */
		/*
		fprintf (stderr, "\nAdding %g\n", rsptm);
		mgr_->printBins();
		fprintf (stderr, "50th: %d  60th: %d  70th: %d\n",
			 mgr_->getPercentile(0.5), 
			 mgr_->getPercentile(0.6),
			 mgr_->getPercentile(0.7));
		*/

		// dump output
		FILE* fp = mgr_->get_outfp();
		char* srv_nodeaddr = 
			Address::instance().print_nodeaddr(agent_->daddr());
		char* srv_portaddr = 
			Address::instance().print_nodeaddr(agent_->dport());
		if (fp) {
			fprintf (fp, "%-11.6f %-10d %-10d %-10.3f %s.%-6s %-7d\n",
				 now, reqsize_, rspsize_, 
				 rsptm,
				 srv_nodeaddr, srv_portaddr,
				 mgr_->get_active());
			fflush (fp);
		}

		fp = mgr_->get_statsfp();
		FullTcpAgent* tcp = (FullTcpAgent*) agent_;
		FullTcpAgent* srv_tcp = (FullTcpAgent*) get_server()->get_agent();
		char* cli_nodeaddr = 
			Address::instance().print_nodeaddr(agent_->addr());
		char* cli_portaddr = 
			Address::instance().print_nodeaddr(agent_->port());
		if (fp) {
			fprintf (fp, "%-11.6f %s %-10s %-7d %-7d %s %-10s %-7d %-7d\n",
				 now, cli_nodeaddr, cli_portaddr,
				 tcp->get_nrexmitpack(), 
				 tcp->get_nrexmit(),
				 srv_nodeaddr, srv_portaddr,
				 srv_tcp->get_nrexmitpack(), 
				 srv_tcp->get_nrexmit());
			fflush (fp);
		}
	}
}


/*:::::::::::::::::::::::::: WLANTRAFSERVERAPP ::::::::::::::::::::::::::::::*/

static class WLANTrafFullServerAppClass : public TclClass {
public:
	WLANTrafFullServerAppClass() : TclClass("Application/WLANTrafFull/Server") {}
	TclObject* create(int, const char*const*) {
		return (new WLANTrafFullServerApp);
	}
} class_app_wlantraf_server;

WLANTrafFullServerApp::~WLANTrafFullServerApp()
{
	Tcl& tcl = Tcl::instance();
	if (agent_ != NULL) {
		tcl.evalf ("delete %s", agent_->name());
	}
}

void WLANTrafFullServerApp::send()
{
	if (!running_) {
		return;
	}

	if (mgr_->debug() > 1) {
		fprintf(stderr,"server %s (%d)> sent %d-byte response at %f\n",
			name(), id_, rspsize_, mgr_->now());
	}
	
	// send response
	agent_->sendmsg(rspsize_, "MSG_EOF");
	stop();
}

void WLANTrafFullServerApp::stop()
{
	running_ = 0;
}

void WLANTrafFullServerApp::recycle()
{
	id_ = 0;
	reqsize_ = 0;
	rspsize_ = 0;
	totalbytes_ = 0;
	agent_ = NULL;
}


void WLANTrafFullServerApp::recv(int bytes)
{
	if (!running_) {
		return;
	}

	totalbytes_ += bytes;

	// we've received a packet
	if (mgr_->debug() > 3)
		fprintf (stderr, "server %s (%d)> received %d bytes at %f\n", 
			 name(), id_, bytes, mgr_->now());

	if (totalbytes_ == reqsize_) {
		// send data
		send();
	}
}
