//* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */

/* 
 * Copyright 2006, Clemson University
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *    1. Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Based on data from:
 * Meng et al, "Characterizing Flows in Large Wireless Data Networks", 
 * MobiCom 2004
 * AND
 * Kotz and Essien, "Analysis of a Campus-wide Wireless Network",
 * Mobicom 2002
 */

#include <tclcl.h>
#include "lib/bsd-list.h"
#include "random.h"
#include "tcp.h"
#include "tcp-sink.h"
#include "wlantraf.h"

/*::::::::::::::::::::::::::::::: WLANTRAF::::::::::::::::::::::::::::::::::*/

static class WLANTrafClass : public TclClass {
public:
	WLANTrafClass() : TclClass("WLANTraf") {}
	TclObject* create(int, const char*const*) {
		return (new WLANTraf);
	}
} class_wlantraf;

WLANTraf::WLANTraf() :
	TclObject(), timer_(this), connection_interval_(0), 
	client_ind_(0), client_nodes_(0), delayed_acks_(1),
	outfp_(NULL), ID_(-1), run_(0), debug_(0),
	incoming_percent_(50), outgoing_percent_(50),
	active_connections_(0), total_connections_(-1), running_(0),
	flowarrive_rv_(NULL), flowsize_rv_(NULL)
{
	int i;
	strcpy (tcptype_, "Sack1");
	server_ = NULL;
	for (i=0; i<MAX_NODES; i++) {
		client_[i] = NULL;
	}
}

WLANTraf::~WLANTraf() {}

TcpAgent* WLANTraf::picktcp()
{
	TcpAgent* a;
	Tcl& tcl = Tcl::instance();

	tcl.evalf ("%s alloc-tcp %s", name(), tcptype_);
	a = (TcpAgent*) lookup_obj (tcl.result());
	if (a == NULL) {
		fprintf (stderr, "Failed to allocate a TCP agent\n");
		abort();
	}
	if (debug_ > 1) {
		fprintf (stderr, 
			 "\tflow %d created new TCPAgent %s\n",
			 total_connections_, a->name());
	}

	return a;
}

TcpSink* WLANTraf::picksink()
{
	TcpSink* a;
	Tcl& tcl = Tcl::instance();

	tcl.evalf ("%s alloc-sink %d", name(), delayed_acks_);
	a = (TcpSink*) lookup_obj (tcl.result());
	if (a == NULL) {
		fprintf (stderr, "Failed to allocate a TCP sink\n");
		abort();
	}
	if (debug_ > 1) {
		fprintf (stderr, 
			 "\tflow %d created new TCPSink %s\n",
			 total_connections_, a->name());
	}

	return a;
}


WLANTrafServerApp* WLANTraf::pickServerApp()
{
	WLANTrafServerApp* a;

	if (serverAppPool_.empty()) {
		Tcl& tcl = Tcl::instance();
		tcl.evalf ("%s alloc-server-app", name());
		a = (WLANTrafServerApp*) lookup_obj (tcl.result());
		if (a == NULL) {
			fprintf (stderr, 
				"Failed to allocate a WLANTraf server app\n");
			abort();
		}
		if (debug_ > 1)
			fprintf (stderr, "\tflow %d created new ",
				 total_connections_);
	} else {
		a = serverAppPool_.front();   // grab top of the queue
		serverAppPool_.pop();         // remove top from queue

		if (debug_ > 1)
			fprintf (stderr, "\tflow %d got ", 
				 total_connections_);
	}

	// initialize server app
	a->set_id(total_connections_);
	a->set_mgr(this);

	if (debug_ > 1)
		fprintf (stderr, "ServerApp %s (%d in pool, %d active)\n",
			 a->name(), serverAppPool_.size(), 
			 serverAppActive_.size()+1);

	return a;
}

WLANTrafClientApp* WLANTraf::pickClientApp()
{
	WLANTrafClientApp* a;

	if (clientAppPool_.empty()) {
		Tcl& tcl = Tcl::instance();
		tcl.evalf ("%s alloc-client-app", name());
		a = (WLANTrafClientApp*) lookup_obj (tcl.result());
		if (a == NULL) {
			fprintf (stderr, 
				"Failed to allocate a WLANTraf client app\n");
			abort();
		}
		if (debug_ > 1)
			fprintf (stderr, "\tflow %d created new ",
				 total_connections_);
	} else {
		a = clientAppPool_.front();   // grab top of the queue
		clientAppPool_.pop();         // remove top from queue

		if (debug_ > 1)
			fprintf (stderr, "\tflow %d got ", 
				 total_connections_);
	}

	// initialize client app
	a->set_id(total_connections_);
	a->set_mgr(this);

	if (debug_ > 1)
		fprintf (stderr, "ClientApp %s (%d in pool, %d active)\n",
			 a->name(), clientAppPool_.size(), 
			 clientAppActive_.size()+1);

	return a;
}


void WLANTraf::recycle(WLANTrafClientApp* app)
{
	if (app == NULL)
		return;

	// find the client app in the active pool
	map<string, WLANTrafClientApp*>::iterator ca_iter = 
		clientAppActive_.find(app->get_agent_name());
	if (ca_iter == clientAppActive_.end()) 
		return;

	// remove the client app from the active pool
	clientAppActive_.erase(ca_iter);

	// insert the client app into the inactive pool
	clientAppPool_.push (app);

	if (debug_ > 2) {
		fprintf (stderr, "\tClientApp %s (%d) moved to pool ", 
			 app->name(), app->get_id());
		fprintf (stderr, "(%d in pool, %d active)\n", 
			 clientAppPool_.size(), clientAppActive_.size());
	}

	// recycle app
	app->recycle();
}

void WLANTraf::recycle(WLANTrafServerApp* app)
{
	if (app == NULL)
		return;

	// find the server app in the active pool
	map<string, WLANTrafServerApp*>::iterator sa_iter = 
		serverAppActive_.find(app->get_agent_name());
	if (sa_iter == serverAppActive_.end()) 
		return;

	// remove the server app from the active pool
	serverAppActive_.erase(sa_iter);

	// insert the server app into the inactive pool
	serverAppPool_.push (app);

	if (debug_ > 2) {
		fprintf (stderr, "\tServerApp %s (%d) moved to pool ",
			 app->name(), app->get_id());
		fprintf (stderr, "(%d in pool, %d active)\n", 
			 serverAppPool_.size(), serverAppActive_.size());
	}

	// recycle app
	app->recycle();
}

double WLANTraf::connection_interval() 
{
	return connection_interval_;
}

int WLANTraf::get_flowsize() 
{
	return (int) (flowsize_rv_->value());
}

void WLANTraf::setup_connection()
/*
 * Setup a new connection, including creation of Agents and Apps
 */
{
	Tcl& tcl = Tcl::instance();

	// incr count of connections
	active_connections_++;
	total_connections_++;

	if (debug_ > 1) {
		fprintf (stderr, 
		 "\nWLANTraf %s> new flow %d  total active: %d at %f\n",
		 name(), total_connections_, active_connections_, now());
	}

	// pick tcp agent for client and server
	TcpAgent* c_tcp = picktcp();
	TcpSink* s_sink = picksink();
	TcpAgent* s_tcp = picktcp();
	TcpSink* c_sink = picksink();

	// rotate through nodes assigning connections
	client_ind_++;
	if (client_ind_ >= client_nodes_)
		client_ind_ = 0;

	// attach agents to nodes (server_ client_)
	tcl.evalf ("%s attach %s", server_->name(), s_tcp->name());
	tcl.evalf ("%s attach %s", client_[client_ind_]->name(), 
		   c_tcp->name());
	tcl.evalf ("%s attach %s", server_->name(), s_sink->name());
	tcl.evalf ("%s attach %s", client_[client_ind_]->name(), 
		   c_sink->name());

	// create WLANTrafApps
	WLANTrafClientApp* client_app = pickClientApp();
	WLANTrafServerApp* server_app = pickServerApp();

	// set TCP options
	tcl.evalf ("%s setup-tcp %s %s %s %s %d 1", name(), c_tcp->name(), 
		   s_sink->name(), client_app->name(), server_app->name(), total_connections_);
	tcl.evalf ("%s setup-tcp %s %s %s %s %d 0", name(), s_tcp->name(), 
		   s_sink->name(), client_app->name(), server_app->name(), total_connections_);

	// setup connection between client and server
	tcl.evalf ("set ns [Simulator instance]");
	tcl.evalf ("$ns connect %s %s", c_tcp->name(), s_sink->name());
	tcl.evalf ("%s listen", s_sink->name());
	tcl.evalf ("$ns connect %s %s", s_tcp->name(), c_sink->name());
	tcl.evalf ("%s listen", c_sink->name());

	// attach TCPs to WLANTrafApps
	c_tcp->attachApp ((Application*) client_app);
	s_tcp->attachApp ((Application*) server_app);
	client_app->set_server(server_app);
	client_app->set_agent(c_tcp);
	server_app->set_agent(s_tcp);

	// put apps in active list
	clientAppActive_[c_tcp->name()] = client_app;
	serverAppActive_[s_tcp->name()] = server_app;

	// start WLANTrafApps
	client_app->start();
	//	server_app->start();

	// set time for next connection to start
	connection_interval_ = flowarrive_rv_->value();
}

void WLANTraf::start()
{            
	running_ = 1;

	// schedule first connection
	timer_.sched (0);
}

void WLANTraf::stop()
{
	running_ = 0;

	// output stats
	if (debug_ > 0) {
		fprintf (stderr, "total connections created: %d ", 
			 total_connections_+1);
		fprintf (stderr, "(in pool: %d  active: %d)\n", 
			 serverAppPool_.size(), serverAppActive_.size());
	}
	
	// delete timer
	timer_.force_cancel();

	// close output file
	if (outfp_)
		fclose(outfp_);
}

int WLANTraf::command(int argc, const char*const* argv) {
	if (argc == 2) {
		if (strcmp (argv[1], "start") == 0) {
			start();
			return (TCL_OK);
		}
		else if (strcmp (argv[1], "stop") == 0) {
			stop();
			return (TCL_OK);
		}
	}
	else if (argc == 3) {
		if (!strcmp (argv[1], "set-client")) {
			if (client_ind_ >= MAX_NODES) {
				return (TCL_ERROR);
			}
			client_[client_ind_] = (Node*) 
				lookup_obj(argv[2]);
			if (client_[client_ind_] == NULL) {
				return (TCL_ERROR);
			}
			client_ind_++;
			client_nodes_++;

			return (TCL_OK);			
		}
		else if (!strcmp (argv[1], "set-server")) {
			server_ = (Node*) lookup_obj(argv[2]);
			if (server_ == NULL)
				return (TCL_ERROR);
			return (TCL_OK);			
		}
		else if (!strcmp (argv[1], "set-flow_size")) {
                        int res = lookup_rv (flowsize_rv_, argv[2]);
                        if (res == TCL_ERROR) {
                                delete flowsize_rv_;
                                fprintf (stderr, "Invalid flow size ");
                                fprintf (stderr, "random variable\n");
                                return (TCL_ERROR);
                        }
                        return (TCL_OK);
                }
                else if (!strcmp (argv[1], "set-flow_arrive")) {
                        int res = lookup_rv (flowarrive_rv_, argv[2]);
                        if (res == TCL_ERROR) {
                                delete flowarrive_rv_;
                                fprintf (stderr,"Invalid flow arrive ");
                                fprintf (stderr, "random variable\n");
                                return (TCL_ERROR);
                        }
                        return (TCL_OK);
                }
		else if (!strcmp (argv[1], "set-tcp-type")) {
			strcpy (tcptype_, argv[2]);
			return (TCL_OK);
		}
		else if (!strcmp (argv[1], "set-delayed-acks")) {
			delayed_acks_ = atoi (argv[2]);
			return (TCL_OK);
		}
		else if (!strcmp (argv[1], "set-incoming-percent")) {
			incoming_percent_ = atoi(argv[2]);
			return (TCL_OK);
		}
		else if (!strcmp (argv[1], "set-outgoing-percent")) {
			outgoing_percent_ = atoi(argv[2]);
			return (TCL_OK);
		}
		else if (strcmp (argv[1], "set-outfile") == 0) {
			outfp_ = fopen (argv[2], "w");
			if (outfp_)
				return (TCL_OK);
			else 
				return (TCL_ERROR);
		}
		else if (strcmp (argv[1], "set-ID") == 0) {
			ID_ = (int) atoi (argv[2]);
			return (TCL_OK);
		}
		else if (strcmp (argv[1], "set-run") == 0) {
			run_ = (int) atoi (argv[2]);
			return (TCL_OK);
		}
		else if (strcmp (argv[1], "set-debug") == 0) {
			debug_ = (int) atoi (argv[2]);
			return (TCL_OK);
		}
		else if (!strcmp (argv[1], "start-server")) {
			WLANTrafServerApp* app = (WLANTrafServerApp*) lookup_obj(argv[2]);
			if (app == NULL) { 
				return (TCL_ERROR);
			}
			app->start();
			return (TCL_OK);
		}
		else if (!strcmp (argv[1], "server-done")) {
			WLANTrafServerApp* app = (WLANTrafServerApp*) lookup_obj(argv[2]);
			if (app == NULL) { 
				return (TCL_ERROR);
			}
			app->dump_output();
			return (TCL_OK);
		}
		else if (strcmp (argv[1], "recycle") == 0) {
			WLANTrafClientApp* ca = (WLANTrafClientApp*) lookup_obj(argv[2]);
			WLANTrafServerApp* sa = ca->get_server();

			if (debug_ > 1)
			      fprintf (stderr, "client %s (%d)> DONE at %f\n",
					 ca->name(), ca->get_id(), now());

			// recycle apps
			recycle (ca);
			recycle (sa);

			active_connections_--;   // one less active conn
			return (TCL_OK);
		}
	}
	return TclObject::command(argc, argv);
}


/*:::::::::::::::::::::::::::: WLANTRAFTIMER ::::::::::::::::::::::::::::::::*/

void WLANTrafTimer::expire(Event* = 0) 
{

}

void WLANTrafTimer::handle(Event* e)
{
	if (!mgr_->running()) {
		if (mgr_->get_active() != 0) {
			TimerHandler::handle(e);
			sched (1);              // check back in 1 second...
		}
		return;
	}

	// setup new connection
	mgr_->setup_connection();

	TimerHandler::handle(e);

	// schedule time for next connection
	if (mgr_->debug() > 1) {
		fprintf (stderr, "\tnext connection scheduled in %f s...\n", 
			 mgr_->connection_interval());
	}
	sched (mgr_->connection_interval());
}


/*:::::::::::::::::::::::::: WLANTRAFCLIENTAPP ::::::::::::::::::::::::::::::*/

static class WLANTrafClientAppClass : public TclClass {
public:
	WLANTrafClientAppClass() : TclClass("Application/WLANTraf/Client") {}
	TclObject* create(int, const char*const*) {
		return (new WLANTrafClientApp);
	}
} class_app_wlantraf_client;

WLANTrafClientApp::~WLANTrafClientApp()
{
	Tcl& tcl = Tcl::instance();
	if (agent_ != NULL) {
		tcl.evalf ("delete %s", agent_->name());
	}
}

void WLANTrafClientApp::start()
{
	running_ = 1;
	send();
}

void WLANTrafClientApp::stop()
{
	running_ = 0;
}

void WLANTrafClientApp::recycle()
{
	//	Tcl& tcl = Tcl::instance();
	id_ = 0;
	rspsize_ = 0;
	/*
	if (agent_ != NULL) {
		tcl.evalf ("delete %s", agent_->name());
	}
	*/
	agent_ = NULL;
	server_ = NULL;
}


void WLANTrafClientApp::send()
{
	int flowsize;

	if (!running_) {
		return;
	}

	// get flow size
	flowsize = mgr_->get_flowsize();
	reqsize_ = (int) (flowsize * (mgr_->get_outgoing_percent() / 100.0));
	rspsize_ = (int) (flowsize * (mgr_->get_incoming_percent() / 100.0));
	if (reqsize_ < 1) {
		reqsize_ = 1;
	}
	if (rspsize_ < 1) {
		rspsize_ = 1;
	}
	if (mgr_->debug() > 2) {
		fprintf (stderr, "client %s (%d)> picked flowsize %d  req %d  rsp %d\n", 
			 name(), id_, flowsize, reqsize_, rspsize_);
	}

	time_of_req_ = mgr_->now();

	server_->set_reqsize(reqsize_);
	server_->set_rspsize(rspsize_);
	server_->set_timeofreq(time_of_req_);
	
	// send request
	agent_->sendmsg(reqsize_);
	if (mgr_->debug() > 1)
		fprintf (stderr, "client %s (%d)> sent %d-byte req (%d-byte rsp) at %f\n",
			 name(), id_, reqsize_, rspsize_, time_of_req_);

	agent_->close();
	stop();
}



/*:::::::::::::::::::::::::: WLANTRAFSERVERAPP ::::::::::::::::::::::::::::::*/

static class WLANTrafServerAppClass : public TclClass {
public:
	WLANTrafServerAppClass() : TclClass("Application/WLANTraf/Server") {}
	TclObject* create(int, const char*const*) {
		return (new WLANTrafServerApp);
	}
} class_app_wlantraf_server;

WLANTrafServerApp::~WLANTrafServerApp()
{
	Tcl& tcl = Tcl::instance();
	if (agent_ != NULL) {
		tcl.evalf ("delete %s", agent_->name());
	}
}

void WLANTrafServerApp::start() 
{
	running_ = 1;
	send();
}

void WLANTrafServerApp::dump_output()
{
	// dump output
	double now = mgr_->now();
	FILE* fp = mgr_->get_outfp();
	char* dst_nodeaddr = 
		Address::instance().print_nodeaddr(agent_->daddr());
	char* dst_portaddr = 
		Address::instance().print_nodeaddr(agent_->dport());
	if (fp) {
		fprintf (fp, "%-11.6f %-10d %-10d %-10.3f %-5s %-7s %-7d\n",
			 now, reqsize_, rspsize_, 
			 (now - time_of_req_) * 1000.0, 
			 dst_nodeaddr, dst_portaddr,
			 mgr_->get_active());
		fflush (fp);
	}
}

void WLANTrafServerApp::send()
{
	if (!running_) {
		return;
	}

	// send response
	agent_->sendmsg(rspsize_);
	if (mgr_->debug() > 1) {
		fprintf(stderr,"server %s (%d)> sent %d-byte response at %f\n",
			name(), id_, rspsize_, mgr_->now());
	}
	agent_->close();
	stop();
}

void WLANTrafServerApp::stop()
{
	running_ = 0;
}

void WLANTrafServerApp::recycle()
{
	//	Tcl& tcl = Tcl::instance();
	id_ = 0;
	reqsize_ = 0;
	rspsize_ = 0;
	time_of_req_ = 0;
	/*
	if (agent_ != NULL) {
		tcl.evalf ("delete %s", agent_->name());
	}
	*/
	agent_ = NULL;
}


