/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */

/* 
 * Copyright 2006, Clemson University
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *    1. Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Based on data from:
 * Meng et al, "Characterizing Flows in Large Wireless Data Networks", 
 * MobiCom 2004
 * AND
 * Kotz and Essien, "Analysis of a Campus-wide Wireless Network",
 * Mobicom 2002
 */

#define MAX_NODES 10

#ifndef ns_wlantraf_h
#define ns_wlantraf_h

#include "timer-handler.h"
#include "app.h"
#include "node.h"
#include "ranvar.h"
#include <string>
#include <queue>
#include <map>

class TcpAgent;
class RandomVariable;
class WLANTraf;
class WLANTrafTimer;
class WLANTrafServerApp;
class WLANTrafClientApp;

/*:::::::::::::::::::::::::: TIMER HANDLER class ::::::::::::::::::::::::::::*/

class WLANTrafTimer : public TimerHandler {
public:
	WLANTrafTimer(WLANTraf* mgr) : TimerHandler(), mgr_(mgr) {}
	inline WLANTraf* mgr() {return mgr_;}
protected:
	virtual void handle(Event* e);
	virtual void expire(Event* e);
	WLANTraf* mgr_;	                    // pointer to WLANTraf object
};

/*:::::::::::::::::::::::: WLANTRAF APPLICATION Classes ::::::::::::::::::::*/

class WLANTrafClientApp : public Application {
 public:
	WLANTrafClientApp() : Application(), id_(0), running_(0), 
			      totalbytes_(0), reqsize_(0), rspsize_(0), 
                              time_of_req_(0.0), server_(NULL), mgr_(NULL) {};
	~WLANTrafClientApp();
	void start();
	void stop();
	void recycle();

	inline void set_server(WLANTrafServerApp* server) {server_ = server;}
	inline WLANTrafServerApp* get_server() {return server_;}
	inline void set_agent(Agent* tcp) {agent_ = tcp;}
	inline const char* get_agent_name() {return agent_->name();}
	inline void set_mgr(WLANTraf* mgr) {mgr_ = mgr;}
	inline void set_id (int id) {id_ = id;}
	inline int get_id () {return id_;}
	inline void delete_agent() {delete agent_;}

 protected:
	void send();

	int id_;
	int running_;
	int totalbytes_;                // total bytes received so far
	int reqsize_;
	int rspsize_;
	double time_of_req_;
	WLANTrafServerApp* server_;     // pointer to Server
	WLANTraf* mgr_;                 // pointer to WLANTraf object
};


class WLANTrafServerApp : public Application {
 public:
	WLANTrafServerApp() : Application(), id_(0), running_(0), 
			      reqsize_(0), rspsize_(0), 
			      time_of_req_(0.0), mgr_(NULL) {};
	~WLANTrafServerApp();
	void stop();
	inline const char* get_agent_name() {return agent_->name();}
	inline void set_agent( Agent* tcp) {agent_ = tcp;}
	inline void set_mgr(WLANTraf* mgr) {mgr_ = mgr;}
	inline void set_id (int id) {id_ = id;}
	inline int get_id () {return id_;}
	inline void set_reqsize(int size) {reqsize_ = size;}
	inline void set_rspsize(int size) {rspsize_ = size;}
	inline void set_timeofreq (double time) {time_of_req_ = time;}
	inline void delete_agent() {delete agent_;}
	void recycle();
	void start();
	void dump_output();

 protected:
	void send();

	int id_;
	int running_;
	int reqsize_;
	int rspsize_;
	double time_of_req_;

	WLANTraf* mgr_;               // pointer to WLANTraf object
};

/*::::::::::::::::::::::::: class WLANTRAF :::::::::::::::::::::::::::::::::*/

class WLANTraf : public TclObject {
 public:
	WLANTraf();
	~WLANTraf();
	void recycle (WLANTrafClientApp*);
	void recycle (WLANTrafServerApp*);
	void setup_connection ();

	inline double now() {return Scheduler::instance().clock();}
	inline int get_active() {return active_connections_;}
	inline int get_total() {return total_connections_;}
	inline int running() {return running_;}
	inline int debug() {return debug_;}
	inline int get_ID() {return ID_;}
	inline int get_outgoing_percent() { return outgoing_percent_;}
	inline int get_incoming_percent() { return incoming_percent_;}

	double connection_interval();
	int get_flowsize();

	inline FILE* get_outfp() {return outfp_;}

 protected:
	virtual int command (int argc, const char*const* argv);
	void start();
	void stop();
	void start_server (WLANTrafServerApp*);
	void dump_output (WLANTrafServerApp*);

	TcpAgent* picktcp();
	TcpSink* picksink();
	WLANTrafServerApp* pickServerApp();
	WLANTrafClientApp* pickClientApp();	

	WLANTrafTimer timer_;
	double connection_interval_;  // set in setup_connection()

	// TCL configurable variables
	Node* server_;
	Node* client_[MAX_NODES];
	int client_ind_;           // index into client_ array
	int client_nodes_;         // number of wireless client nodes
	char tcptype_[20];         // {Reno, Tahoe, Newreno, Sack}
	int delayed_acks_;         // use delayed acks
	FILE* outfp_;
	int ID_;                   // WLANTraf cloud ID
	int run_;                  // exp run number (for RNG stream selection)
	int debug_;
	int incoming_percent_;     // percent of flow_size wired2wireless
	int outgoing_percent_;     // percent of flow_size wireless2wired

	int active_connections_;   // number of active connections
	int total_connections_;    // number of total connections
	int running_;              // start new connections?
	
	// statistics objects
	RandomVariable* flowarrive_rv_;
	RandomVariable* flowsize_rv_;

	// helper methods
	TclObject* lookup_obj(const char* name) {
                TclObject* obj = Tcl::instance().lookup(name);
                if (obj == NULL) 
                        fprintf(stderr, "Bad object name %s\n", name);
                return obj;
        }

	inline int lookup_rv (RandomVariable*& rv, const char* name) {
		if (rv != NULL)
			Tcl::instance().evalf ("delete %s", rv->name());
		rv = (RandomVariable*) lookup_obj (name);
		return rv ? (TCL_OK) : (TCL_ERROR);
	}

	// Agent and App Pools	
	queue<WLANTrafClientApp*> clientAppPool_;
	queue<WLANTrafServerApp*> serverAppPool_;

	// string = tcpAgent's name
	map<string, WLANTrafClientApp*> clientAppActive_;
	map<string, WLANTrafServerApp*> serverAppActive_;
};

#endif

